/* ifile - intelligent mail filter for EXMH/MH
   ifile is Copyright (C) 1997  Jason Rennie <jr6b+@andrew.cmu.edu>
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program (see file 'COPYING'); if not, write to the Free
   Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA  02111-1307, USA.
   */

/* References:

   [1] - _Unix_System_Programming_Using_C++_ written by Terrence Chan.
         Copyright (C) 1997 by Prentice Hall PTR.
   */

#include <stdarg.h>     /* va_*() */
#include <assert.h>     /* assert() */
#include <string.h>     /* strtok()  strchr() */
#include <stdlib.h>     /* malloc()  free() */
#include <stdio.h>      /* printf() */
#include <sys/stat.h>   /* stat() */
#include <sys/time.h>   /* clock() */
#include <unistd.h>     /* stat() */
#include <sys/types.h>  /* waitpid() WIFEXITED WEXITSTATUS */
#include <sys/wait.h>   /* waitpid() WIFEXITED WEXITSTATUS */
#include <errno.h>      /* required by SunOS 5.5.1/gcc-2.7.2 for errno var */
#include <ctype.h>      /* tolower() */
#include <math.h>       /* log() */

#include "ifile.h"      /* main ifile function library */
#include "error.h"      /* message and error reporting library */

extern FILE * INFO;

/* Global variables used for storing information about documents, idata file */
extern assoc_array_int as_n,       /* number of total words in a category */
                       as_docs,    /* number of documents in a category */
                       as_age,     /* age of a word */
                       as_nk,      /* instances of a word in a specific
				      category */
                       as_sum_nk,  /* total instances of a word */
                       as_index,   /* index of word in ea_word array - 0
				      indicates non-existence */
                       as_msg_freq; /* frequency of word in input document */
extern extendable_array ea_word,  /* list of unique words in the database */
                        ea_msg_words, /* set of unique words in message */
                        ea_category,  /* list of unique categories/maiboxes */
                        ea_lines;     /* the message broken into lines */
extern int num_categories;   /* total number of distinct categories */
extern int num_words;        /* total number of distinct words */
extern int num_docs;         /* total number of documents */
extern int num_msg_words;    /* total number of words in message - repeats of
				a word ARE counted in this number */

/* Global variables for keeping track of time/speed of ifile */
extern clock_t DMZ_start, DMZ_end;

/* Global variable to keep track of mh_profile settings */
extern mh_profile_struct mh_profile;

extern char * dest_folder;  /* name of destination foler */

/* Given a printf style format string and an arbitrarily long list of
   arguments which are in accordance with the format string, ifile_sprintf
   will allocate memory for and create a string according to the given
   information.
   */
/* Written by Jason Rennie <jr6b+@andrew.cmu.edu> for ifile.
   Portions adapted from reference [1], page 109 */
char *
ifile_sprintf (char * format, ...)
{
  char buf[MAX_STR_LEN] = ""; /* holds string to be returned by function */
  va_list ap;
  char * rtn;

  va_start(ap, format);
  assert(vsprintf(buf, format, ap) < MAX_STR_LEN - 1);
  va_end(ap);

  rtn = malloc(strlen(buf)+1);
  strcpy(rtn, buf);
  return rtn;
}


/* Returns a string which is the concatenation of an arbitrary number
   of strings passed as arguments to the function.
   First argumet passed to function is the number of strings passed to
   the function which are to be concatenated.
   */
/* written by Jason Rennie <jr6b+@andrew.cmu.edu> for ifile */
char *
ifile_cats (int num_strings, ...)
{
  va_list ap;
  int string_size = 0;
  char * new_string = NULL;
  int i = 0;

  va_start(ap, num_strings);
  for (i=0; i < num_strings; i++)
    string_size += strlen(va_arg(ap, char *));
  va_end(ap);

  new_string = malloc(string_size+1);
  new_string[0] = '\0';

  va_start(ap, num_strings);
  for (i=0; i < num_strings; i++)
    strcat(new_string, va_arg(ap, char *));
  va_end(ap);

  return new_string;
}


/* Returns 1 if FILE_NAME is executable, 0 otherwise */
/* Written by Jason Rennie <jr6b+@andrew.cmu.edu> for ifile */
int
ifile_exe (char * file_name)
{
  struct stat * tmp_file_stat = (struct stat *) malloc(sizeof(struct stat));
  
  if (!(stat(MH_REFILE, tmp_file_stat) == 0))
      ifile_error("%s does not stat!", MH_REFILE);

  /* checks for executable bit */
  if ((tmp_file_stat->st_mode & S_IXOTH) == 0)
    {
      ifile_verbosify(ifile_quiet, "%s is not executable!  Error# %d\n",
		      file_name, errno);
      return 0;
    } else {
      ifile_verbosify(ifile_verbose, "%s is executable.\n", file_name);
      return 1;
    }
}


/* Given an integer value, allocates space for and returns a string
   representing the integer in character form
   */
/* Written by Jason Rennie <jr6b+@andrew.cmu.edu> for ifile */
char *
itoa (int number)
{  
  linked_list * list = NULL;
  linked_list * new_digit = NULL;
  linked_list * list_ptr = NULL, * old_list_ptr = NULL;
  char buf[MAX_STR_LEN];
  int negative = FALSE;
  int i = 0;
  char * tmp = NULL;

  if (number == 0)
    { 
      tmp = malloc(strlen("0") + 1);
      strcpy(tmp, "0"); 
      return tmp;
    }
  
  if (number < 0)
    {
      number = abs(number);
      negative = TRUE;
    }

  /* Loads digits into stack-like linked list */

  while (number > 0)
    {
      new_digit = (linked_list *) malloc(sizeof(linked_list));
      new_digit->next = list;
      new_digit->digit = number - ((number/10)*10);
      number /= 10;
       list = new_digit;
    }
  
  /* Removes digits from list, copying them into the string to be returned */

  i = 0;
  if (negative == TRUE)
    { 
      buf[0] = '-';
      i++;
    }
  list_ptr = list; 
  for (; i < MAX_STR_LEN - 1; i++)
    {
      buf[i] = 48 + list_ptr->digit;
      old_list_ptr = list_ptr;
      list_ptr = list_ptr -> next;
      free(old_list_ptr);
      if (list_ptr == NULL) break;
    }
  buf[i+1] = '\0';
  tmp = malloc(strlen(buf) + 1);
  strcpy(tmp, buf);
  return tmp;
}


/* Reads up to and including the next feedline (\n) and stores the
   string in the STRING buffer.  This function relies on STRING
   being allocated with space for at least MAX_LEN-1 characters.
   Wrapper for fgets().  Returns a pointer to STRING. */
/* Written by Jason Rennie <jr6b+@andrew.cmu.edu> for ifile */
char *
readline (char * string, int max_len, FILE * stream)
{
  int last;
  
  fgets(string, max_len, stream);
  last = strlen(string)-1;
  if (string[last] == '\n')
    string[last] = '\0';
  
  return string;
}


/* Accepts as parameters an EXTENDABLE_ARRAY of LINES, each of which is a
   null-terminated string which does NOT contain a feedline '\n', and
   a string, MBOX, which indicates the destination mailbox.
   Function passes message to MH rcvstore and returns 0 upon safe execution,
   -1 otherwise. */
/* Written by Jason Rennie <jr6b+@andrew.cmu.edu> for ifile */
int 
ifile_rcvstore (extendable_array lines, char * mbox)
{
  int fifo[2], status;  /* Pipe file descriptors */
  char * mboxplus = NULL;
  int i;
  pid_t child_pid;
  char * buf = NULL;
  char * ifile_header = NULL;
  int buf_siz;
  
  mboxplus = ifile_sprintf("+%s", mbox);
  
  ifile_verbosify(ifile_progress, "Rcvstoring in %s\n", mbox);
  
  if(pipe(fifo)==-1)
    ifile_error("Could not open pipe to rcvstore.");
  
  switch(child_pid = fork())
    {
    case -1:   /* Error */
      ifile_error("Fork to rcvstore failed.\n");
    case 0:    /* Child */
      if(close(fifo[1])==-1)   /* We don't need the writing end */
	ifile_error("Close of writing end in child failed.");
      dup2(fifo[0], 0);  /* Make stdin refer to the inherited reading
			    end of the pipe. */
      execlp(MH_RCVSTORE, MH_RCVSTORE, mboxplus, NULL);
      ifile_error("execlp to rcvstore failed.\n");
    default:   /* Parent */
      if(close(fifo[0])==-1)  /* We don't need the reading end */
	ifile_error("Close of reading end in parent failed.");

      /* Add header to message indicating that it has been filtered
	 by ifile :) */
      ifile_header = ifile_sprintf("X-filter: ifile v%d.%d.%d => %s\n",
				   IFILE_MAJOR_VERSION_NUMBER,
				   IFILE_MINOR_VERSION_NUMBER,
				   IFILE_TRIFLING_VERSION_NUMBER, mbox);
      buf_siz = strlen(ifile_header);
      write(fifo[1], ifile_header, buf_siz);

      /* pipe message to rcvstore child */
      for(i = 0; i < lines.elems; i++)
	{
	  /* send line plus '\n', malloc extra byte for \0 */
	  buf_siz = strlen(EXT_ARRAY_GET(lines, char *, i)) + 1;
	  buf = malloc(buf_siz+1);
	  strcpy(buf, EXT_ARRAY_GET(lines,char*,i));
	  strcat(buf, "\n");
	  write(fifo[1], buf, buf_siz);
	  free(buf);
	}
      if (close(fifo[1])==-1)
	ifile_error("Close of writing end in parent failed.");

      /* allows exmh to recognize new message, highlight proper folder */
      ifile_verbosify(ifile_verbose, "Waiting for rcvstore fork to return\n");
      if ((waitpid(child_pid, &status, 0) == child_pid) && WIFEXITED(status))
	return WEXITSTATUS(status);
      return -1;
    }
}


/* Closes log file and transfers execution of program to refile program. 
   Function does not exit upon normal execution. */
/* Written by Jason Rennie <jr6b+@andrew.cmu.edu> for ifile */
int 
ifile_mh_refile (int argc, char **argv)
{
  ifile_verbosify(ifile_progress, "Executing MH refile program...\n");

  ifile_close_log();

  argv[0] = MH_REFILE;
  /*  return execve(argv[0], argv, environ); */
  return execv(argv[0], argv);
}


/* Wrapper for standard free() function.  Frees memory and then sets
   pointer equal to NULL. */
/* Written by Jason Rennie <jr6b+@andrew.cmu.edu> for ifile */
void 
ifile_free (void * var)
{
  free(var);
  var = NULL;
}


/* Reads in ACC_FILE, adds to the total counts according to
   the parameters of the function and then writes the accuracies file
   back to disk. */
/* Written by Jason Rennie <jr6b+@andrew.cmu.edu> for ifile */
void 
ifile_update_accuracy (char * acc_file, int filters, int refiles)
{
  FILE * ACC = NULL;
  char tmpstr[MAX_STR_LEN] = "";
  char * tmpstrptr = NULL;
  float accuracy = 0.0;

  ifile_verbosify(ifile_progress, "Opening %s for reading...\n",
		  acc_file);
  ACC = fopen(acc_file, "r");
  if (ACC != NULL) 
    {
      readline(tmpstr, sizeof(tmpstr), ACC);
      tmpstrptr = strtok(tmpstr, " \t=");
      while (tmpstrptr != NULL)
	{
	  if (strcmp(tmpstrptr, "filters") == 0) 
	    filters += atoi(strtok(NULL, " \t="));
	  else if (strcmp(tmpstrptr, "refiles") == 0)
	    refiles += atoi(strtok(NULL, " \t="));

	  tmpstrptr = strtok(NULL, " \t=");
	}
      accuracy = (((float)(filters-refiles)/(float)filters)*10000.0)/100.0;
      fclose(ACC);
      ACC = NULL;
    }
  else
    ifile_verbosify(ifile_progress, "Accuracy file (%s) doesn't seem to exist.\n  Assuming values of zero for filters and refiles\n", acc_file);
  
  ifile_verbosify(ifile_verbose, "Opening %s for writing...\n",
		  acc_file);
  ACC = fopen(acc_file, "w");
  if (ACC != NULL)
    {
      /* reset accuracies file if unusual situation has come arisen */
      if ((refiles > filters) && (refiles > 30))
	{
	  refiles = 0;
	  filters = 0;
	  accuracy = 0.0;
	  ifile_verbosify(ifile_progress, "Negative accuracy after 30 refilings.  Resetting filters and refiles values\n");
	}
      
      fprintf(ACC, "filters = %d  refiles = %d\n", filters,
	      refiles);
      fprintf(ACC, "Accuracy = %.2f %%\n", accuracy);
      fclose(ACC);
    }
  else
    {
      ifile_verbosify(ifile_quiet, "Was not able to write accuracies file (%s)\n", acc_file);
    }

  ifile_verbosify(ifile_progress, "filters = %d  refiles = %d  ",
		  filters, refiles);
  ifile_verbosify(ifile_progress, "accuracy = %.2f %%\n", accuracy);
}


/* Reads in user's .mh_profile and returns a MH_PROFILE struct containing
   information on the mail directory path, the context file name, the current
   folder and the user's inbox name. */
/* Written by Jason Rennie <jr6b+@andrew.cmu.edu> for ifile */
mh_profile_struct 
ifile_read_profile (char * home_dir)
{
  char * mh_profile = NULL;
  FILE * MH = NULL;
  mh_profile_struct rtn_value = {NULL, NULL, NULL, NULL};
  char * context = NULL;
  char tmpstr[MAX_STR_LEN];
  char * tmpstrptr = NULL;

  ifile_verbosify(ifile_progress, "Reading %s for user information\n",
		  MH_PROFILE);
  mh_profile = ifile_sprintf("%s/%s", home_dir, MH_PROFILE);
  MH = fopen(mh_profile, "r");

  if (MH == NULL)
    {
      rtn_value.mail_path = strdup("Mail/");
      context = strdup("context");
      rtn_value.draft_folder = strdup("drafts");
      rtn_value.inbox_folder = strdup("inbox");
    }
  else
    {
      readline(tmpstr, sizeof(tmpstr), MH);
      while (!feof(MH)) {
	tmpstrptr = strtok(tmpstr, " \t");
	if (strcmp(tmpstrptr, "Path:") == 0)
	  {
	    tmpstrptr = strtok(NULL, " \t");
	    rtn_value.mail_path = ifile_sprintf("%s/", tmpstrptr);
	  } 
	else if (strcmp(tmpstrptr, "Draft-Folder:") == 0 ||
		 strcmp(tmpstrptr, "draft-folder:") == 0)
	  {
	    tmpstrptr = strtok(NULL, " \t");
	    rtn_value.draft_folder = ifile_sprintf("%s", tmpstrptr);
	  }
	else if (strcmp(tmpstrptr, "context:") == 0)
	  {
	    tmpstrptr = strtok(NULL, " \t");
	    context = ifile_sprintf("%s", tmpstrptr);
	  }
	else if (strcmp(tmpstrptr, "Inbox:") == 0)
	  {
	    tmpstrptr = strtok(NULL, " \t");
	    rtn_value.inbox_folder = ifile_sprintf("%s", tmpstrptr);
	  }
	  readline(tmpstr, sizeof(tmpstr), MH);
      }
      fclose(MH);
    }

  /* use default values if not found in profile */
  if (context == NULL)
    context = strdup("context");
  if (rtn_value.inbox_folder == NULL)
    rtn_value.inbox_folder = strdup("inbox");
  if (rtn_value.draft_folder == NULL)
    rtn_value.draft_folder = strdup("drafts");
  if (rtn_value.mail_path == NULL)
    rtn_value.mail_path = strdup("Mail/");

  rtn_value.context_file = ifile_sprintf("%s/%s%s", home_dir,
					 rtn_value.mail_path, context);
  ifile_free(context);

  ifile_verbosify(ifile_verbose, "mail path = %s\n", rtn_value.mail_path);
  ifile_verbosify(ifile_verbose, "draft folder = %s\n",
		  rtn_value.draft_folder);
  ifile_verbosify(ifile_verbose, "inbox folder = %s\n",
		  rtn_value.inbox_folder);
  ifile_verbosify(ifile_verbose, "context file = %s\n",
		  rtn_value.context_file);

  return rtn_value;
}

/* Returns TRUE if line contains text which should be lexed and analyzed
   as part of a message.  Otherwise, returns FALSE.
   Currently, function will return FALSE if the first 60 characters
   of the line are not spaces, TRUE otherwise. */
/* Written by Christopher B. Browne <cbbrowne@unicomp.net> for ifile */
/* Minor modifications by Jason Rennie <jr6b+@andrew.cmu.edu> */
int 
ifile_valid_line (char * line)
{
  int i = 0;
  
  /* Go thru the line of text, looking for either a space or the end
     of the string. */
  while ((*line != (char) 32) && (*line != (char) NULL))
    {
      i++;
      line++;
    }
  
  /* Where was the first space? */
  if (i > 60)
    return FALSE;
  else
    return TRUE;
}

/* Accepts a file pointer and reads text from the associated file storing
   each line in the EA_LINES array and storing the multiset of words in the
   EA_MSG_WORDS array.  Additionally, function keeps a associative array of the
   words in the message, AS_MSG_FREQ. */
/* Written by Jason Rennie <jr6b+@andrew.cmu.edu> and others for ifile */
void 
ifile_read_message (FILE * FP)
{
  int i, j, length, old_freq;
  int category_freq;            // sum of frequencies for all words in category
  char nonalpha[256];
  char line[MAX_STR_LEN*4]; 
  char *token = NULL;
  char *wrd = NULL;
  char *token_prefix = "";  /* string prepended to tokens to identify them
			       as special in some way */
  char freq_entry[MAX_STR_LEN]; /* string used for index of as_nk array */

  DMZ_start = clock();
  
  ifile_verbosify(ifile_progress, "Reading message...\n");
  
  /* Initialize the non-alphabetic-character token separator */
  /* '.' and '@' have been included so that e-mail addresses and machine
     names are not broken into pieces */
  j = 0;
  for (i = 1; i <= 255; i++)
    {
      if(!((i >= 'a' && i <= 'z') || (i >= 'A' && i <= 'Z') || (i == '.') ||
	   (i == '@')))
	nonalpha[j++] = i;
    }
  nonalpha[j] = '\0';
  
  i = 0;
  
  /* Read the message into an array for processing */
  readline(line, sizeof(line), FP);
  while (!feof(FP))
    {
      EXT_ARRAY_SET(ea_lines, char*, i, strdup(line));
      i++;
      if (ifile_valid_line(line))
	{
	  /* token prefix - used to identify words in important parts of the
	     message */
	  if (strncmp(line, "Subject:", 8) == 0)
	    token_prefix = "subject~";
	  else if (strncmp(line, "From:", 5) == 0)
	    token_prefix = "from~";
	  else
	    token_prefix = "";

	  /* Split line in words and store them in arrays */
	  token = strtok(line, nonalpha);
	  while (token != NULL)
	    {
	      wrd = token;
	      
	      /* strip sentence-ending periods */
	      length = strlen(wrd);
	      while (wrd[length-1] == '.')
		{
		  wrd[length-1] = '\0';
		  length--;
		}
	      
	      /* Skip tiny words */
	      if (length > 2)
		{
		  /* Convert to lower case */
		  while ((*wrd) != '\0')
		    {
		      (*wrd) = tolower(*wrd);
		      wrd++;
		    }
		  /* attach prefix to token */
		  token = ifile_sprintf("%s%s", token_prefix, token);

		  /* update arrays which concern strictly message */
		  old_freq = assoc_array_int_get(as_msg_freq, token);
		  if (old_freq == 0)
		    {
		      EXT_ARRAY_SET(ea_msg_words, char*, num_msg_words,
				    strdup(token));
		      num_msg_words++;
		    }
		  assoc_array_int_set(as_msg_freq, token, old_freq + 1);

		  /* we should probably have a separate function (called
		     update_idata_arrays maybe :) which takes the message
		     arrays and updates the idata ones */
		  /* update idata arrays */
		  if (dest_folder == NULL)
		    dest_folder = "";
		  old_freq = assoc_array_int_get(as_sum_nk, token);
		  if (old_freq == 0)
		    {
		      EXT_ARRAY_SET(ea_word, char *, num_words, strdup(token));
		      assoc_array_int_set(as_index, token, num_words);
		      num_words++;
		    }
		  assoc_array_int_set(as_sum_nk, token, old_freq+1);
		  category_freq = assoc_array_int_get(as_n, dest_folder);
		  assoc_array_int_set(as_n, dest_folder, category_freq+1);
		  sprintf(freq_entry, "%s,%s", token, dest_folder);
		  old_freq = assoc_array_int_get(as_sum_nk, freq_entry);
		  assoc_array_int_set(as_nk, freq_entry, old_freq+1);

		  ifile_verbosify(ifile_debug, "%s (%d) ", token,
				  assoc_array_int_get(as_msg_freq, token));
		}
	      token = strtok(NULL, nonalpha);
	    }
	}
      readline(line, sizeof(line), FP);
    }
  ifile_verbosify(ifile_debug, "\n");
  DMZ_end = clock();
  ifile_verbosify(ifile_progress,
		  "Finishing reading message.  Elapsed time: %.3f sec\n",
		  ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
}

/* might it be better to have the following functions accept a string as
   input? */

/* Expects a valid file pointer which points to the beginning of the doc
   counts line in an idata file.  The line will be read and information
   from it will stored into global variables.  DATA will be advanced to the
   beginning of the next line */
/* Written by Jason Rennie <jr6b+@andrew.cmu.edu> and others for ifile */
void 
ifile_read_category_doc_counts (FILE * DATA)
{
  char line[MAX_STR_LEN*4];
  char *token = NULL;
  int i;
  int category_docs;   /* total number of documents in one category */

  readline(line, sizeof(line), DATA);
  num_docs = 0;
  token = strtok(line, " \t");
  for (i = 1; token != NULL; i++)
    {
      category_docs = atoi(token);
      assoc_array_int_set(as_docs, 
			  EXT_ARRAY_GET(ea_category, char*, i), 
			  category_docs);
      num_docs += category_docs;
      token = strtok(NULL, " \t");
    }
  if (i <= num_categories)
    ifile_error("Bad data file format - line #3");
}

/* Expects a vaild file pointer which points to the beginning of the word
   counts line in an idata file.  The line will be read and information
   from it will be stored into global variables.  DATA will be advanced to
   the beginning of the next line */
/* Written by Jason Rennie <jr6b+@andrew.cmu.edu> and others for ifile */
void 
ifile_read_category_word_counts (FILE * DATA)
{
  char line[MAX_STR_LEN*4];
  char *token = NULL;
  int i;

  readline(line, sizeof(line), DATA);
  token = strtok(line, " \t");
  for (i = 1; token != NULL; i++)
    {
      assoc_array_int_set(as_n, 
			  EXT_ARRAY_GET(ea_category,char*,i),
			  atoi(token)); 
      token = strtok(NULL, " \t");
    }
  
  if (i <= num_categories)
    ifile_error("Bad data file format - line #2");
}

/* Expects a vaild file pointer which points to the beginning of the category
   names line in an idata file.  The line will be read and information
   from it will be stored into global variables.  DATA will be advanced to
   the beginning of the next line */
/* Written by Jason Rennie <jr6b+@andrew.cmu.edu> and others for ifile */
int 
ifile_read_category_names (FILE * DATA)
{
  char line[MAX_STR_LEN*4];
  char *token = NULL;
  int read_categories = 0;    // number of categories actually read

  num_categories = 0; 

  readline(line, sizeof(line), DATA);
  token = strtok(line, " \t");
  ifile_verbosify(ifile_debug, "category token #1 = %s\n", token);
  while (token != NULL)
    {
      num_categories++;
      read_categories++;
      EXT_ARRAY_SET(ea_category, char *, num_categories, strdup(token));
      token = strtok(NULL, " \t");
      ifile_verbosify(ifile_debug, "category token #%d = %s\n",
		      (num_categories + 1), token);
    }
  return read_categories;
}

/* Using information from global varables, function determines which
   category is most appropriate for the stored message and returns the index
   value for that category */
/* written by Jason Rennie <jr6b+@andrew.cmu.edu> and others for ifile */
int 
ifile_rate_categories (void)
{
  int nkw, freq, best_target = -1, i, j;
  float best_rating = -9999999.9, target_rating;
  float docval, nval;
  char *token = NULL;
  char *wrd = NULL;
  char freq_entry[MAX_STR_LEN];

  DMZ_start = clock();
  ifile_verbosify(ifile_progress, "Computing category ratings...\n");
  
  docval = (float)(num_docs + num_categories);
   
  /* Now do the analysis */
  for (i = 1; i <= num_categories; i++)
    {
      token = EXT_ARRAY_GET(ea_category, char*, i);

      /* We don't want to filter into the draft folder */
      if (strcmp(mh_profile.draft_folder, token) == 0) 
	continue;

      /* constants pulled out to increase efficiency */
      nval = (float)(assoc_array_int_get(as_n, token)+num_words);
      target_rating = - (float)(num_msg_words) * (log(nval));
      ifile_verbosify(ifile_debug,
		      "Evaluate folder: %s nval=%.2f init_target_rating=%.2f\n",
		      token, nval, target_rating);
      for (j = 0; j < num_msg_words; j++)
	{
	  wrd = EXT_ARRAY_GET(ea_msg_words, char*, j);
	  sprintf(freq_entry, "%s,%s", wrd, token);
	  nkw = assoc_array_int_get(as_nk, freq_entry);
	  freq = assoc_array_int_get(as_msg_freq, wrd);
	  ifile_verbosify(ifile_debug, "Word,folder  %s  nk=%d\n",
			  freq_entry, nkw);
	  target_rating += freq * log((float)(nkw+1));
	}
      target_rating += log((float)(assoc_array_int_get(as_docs,token)+1)/
			   docval);
      ifile_verbosify(ifile_debug, "%s given rating of %.2f\n", token,
		      target_rating);

      if ((target_rating > best_rating) || (best_target == -1))
	{
          best_rating = target_rating;
	  best_target = i;
	}
    }
  
  DMZ_end=clock();
  ifile_verbosify(ifile_progress, "Chose %s (%d) as destination folder.  Elapsed time: %.3f sec\n",
		  EXT_ARRAY_GET(ea_category, char *, best_target), 
		  best_target,
		  ((double)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
  return best_target;
}
