/* ifile - intelligent mail filter for EXMH/MH
   ifile is Copyright (C) 1997  Jason Rennie <jr6b+@andrew.cmu.edu>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program (see file 'COPYING'); if not, write to the Free
   Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA  02111-1307, USA.
   */

/*
  ifilter originally written by Jason Rennie <jr6b+@andrew.cmu.edu>
  additions and changes by Diego Zamboni, Chris Browne and others
  */
  
/* 
   $Log: ifilter.c,v $
   Revision 1.3  1997/05/22 02:42:00  cbbrowne
   Massive set of changes.

      Here's a substantial group of changes:
      Christopher Browne, May 21/97;
      cbbrowne@unicomp.net:

      a) Renamed some variables
      - Renamed "tmpstrptr" to "token".  Rather more descriptive, is it
        not?
      - "tmpstr" --> "line".
      - "tmpstr2" --> "freq_entry"
      - "tv" --> "categories"
      - All extendable_arrays are now prefixed with "ea".
      - And associative arrays are now prefixed with "as".

      b) Broke the main program into a number of functions.
      - Broke out a bunch of functions.  read_message(),
        read_ifilter_idata(), rate_categories(), process_frequency_line().

      - Created the new associative array: as_msg_freq.  It gets assigned in
        parallel with ea_words whilst parsing the words in the current
        message.  The point of the exercise is that in
        process_frequency_line(), we return (skipping the line in idata) if
        the word in .idata was not in the message (and thus is not in
        as_msg_freq).

        This is the primary optimization that takes place, and it is really
        rather effective.  It takes run times down (for me) from ~7s to
        ~0.5s per message.  I haven't measured the change in memory
        consumption; that should have improved similarly.

      - Took invariants out of the "category rating" loop.  In particular,
        we can take invariant divisors out of the "log" term, and put them
        in right at the beginning by initializing the target_rating as:
            nval = (float)(assoc_array_int_get(as_n, token)+tot_words);
            target_rating = (float)(word_count) * (log(nval));

      - Fixed a couple of problems where strings needed to be either
        allocated or referenced from the "extendable_array" in order to
        keep an index value for the "associative_array" implementation.

      - I'd quite like to add "." and "@" to the token list.  This allows
        email addresses, newsgroup names, and network IDs to be kept as
        complete tokens.  This ought to improve discrimination somewhat.
        There's the downside that we add additional entries for words that
        end sentences.  Which has the downside that we'd have both "list"
        and "list." on the list of words. Under the "new, more
        RAM-friendly" regime, the cost is merely that some words will be
        repeated twice, which costs not very much. And there is the upside
        that the algorithm could then discriminate between these words.

   */

/* Concerning Chris's desire for '.' and '@' to be added to the set of valid
   characters, they are now part of the set.  I have included in the lexing
   of the message a line to remove any periods at the end of a word.  
   
   I've also moved many of the functions to ifile.c (which I am using as a
   library of functions for ifile).  Any functions which are specific to
   ifilter.c have been left here.

   jr6b+@andrew.cmu.edu */

#include <sys/wait.h>     /* wait() */
#include <math.h>         /* log() */
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>     /* clock() */
#include <stdarg.h>       /* va_*() */
#include <assert.h>       /* assert() */
#include <unistd.h>       /* pipe() fork() close() dup2() execlp() */
#include <errno.h>	  /* required by Solaris 5.5.1/gcc-2.7.2 for errno variable */
#include <string.h>

#include "assoc_array_int.h"

#include "ifile.h"        /* standard ifile library */
#include "error.h"        /* ifile message and error reporting library */
  
/* I've moved all these variables out into the "global" arena.
   Down the road, they should be selectively allocated into individual
   functions as needed 

   cbbrowne@unicomp.net
   */

/* I've added the prefix "as_" to associative arrays, and "ea_" to
   extendable arrays.  Makes it easier to pick them out.  Hungarian
   notation may be evil when applied universally, but is pretty useful
   when applied parsimoniously.

   cbbrowne@unicomp.net
 */

/* Global variables used for storing information about documents, idata file */
assoc_array_int as_n,       /* number of total words in a category */
                as_docs,    /* number of documents in a category */
                as_age,     /* age of a word */
                as_nk,      /* instances of a word in a specific
			       category */
                as_sum_nk,  /* total instances of a word */
                as_index,   /* index of word in ea_word array - 0 indicates
			       non-existence */
                as_msg_freq;   /* frequency of word in input document */
extendable_array ea_word,      /* a list of the unique words in the database */
                 ea_msg_words, /* set of unique words in message */
                 ea_category,  /* a list of the unique categories/maiboxes */
                 ea_lines;     /* the message broken into lines */
int num_categories;   /* total number of distinct categories */
int num_words;        /* total number of distinct words */
int num_docs;         /* total number of documents */
int num_msg_words;    /* total number of distinct words in message */

/* Global variables for keeping track of time/speed of ifile */
clock_t DMZ_start, DMZ_end;

/* Global variable to keep track of mh_profile settings */
mh_profile_struct mh_profile;

char * dest_folder = NULL;  /* name of destination foler */
                            /* used for setting as_n when reading message */

/* ifilter specific function prototypes */
void ifilter_read_idata (FILE * DATA);
int ifilter_process_frequency_line (char *line);
int ifilter_read_word_frequencies (FILE * DATA);
void ifilter_print_usage (void);


/* Main program */
/* written by Jason Rennie <jr6b+@andrew.cmu.edu> and others for ifile */
int 
main(int argc, char **argv)
{
  char *data_file = NULL;
  char *acc_file = NULL;
  char *home_dir = NULL;
  FILE *DATA = NULL;
  int best_target;

  if (argc > 1)
    ifilter_print_usage();
  
  EXT_ARRAY_INIT(ea_word, char *, 300);
  EXT_ARRAY_INIT(ea_msg_words, char *, 300);
  EXT_ARRAY_INIT(ea_category, char *, 50);
  EXT_ARRAY_INIT(ea_lines, char *, 100);
  assoc_array_int_init(as_n);
  assoc_array_int_init(as_docs);
  assoc_array_int_init(as_age);
  assoc_array_int_init(as_nk);
  assoc_array_int_init(as_sum_nk);
  assoc_array_int_init(as_index);
  assoc_array_int_init(as_msg_freq);

  ifile_open_log(argc, argv);
  
  /* Get home directory */
  home_dir = getenv("HOME");
  if (home_dir == NULL)
    ifile_error("Fatal: HOME environment variable not defined!\n");
  ifile_verbosify(ifile_verbose, "home directory = %s\n", home_dir);
  
  /* Reads in info from ~/.mh_profile - currently not used for much */
  mh_profile = ifile_read_profile(home_dir);

  /* Significant change - cbbrowne@unicomp.net, May 20, 1997
     Read/lex the message *before* reading .idata.
     This allows us to skip words that aren't used. */
  
  /* Initialize first; this means that we can call read_message
     multiple times, which makes it reusable in (oh, say) irefile... */
  num_words = 0;

  /* read a message from standard input */
  ifile_read_message(stdin); 
  
  /* If the message was read successfully, then update the accuracy file
     to reflect an additional message filtered */
  acc_file = ifile_sprintf("%s/%s", home_dir, IFILE_ACCURACY);
  ifile_update_accuracy(acc_file, 1, 0);

  /* Now read the idata database */
  data_file = ifile_sprintf("%s/.idata", home_dir);
  DATA = fopen(data_file, "r");
  if (DATA == NULL)
    {
      /* Can't read .idata; let's just store to "inbox", and abort. */
      ifile_rcvstore(ea_lines, "inbox");
      exit(0);
    }
  
  ifile_verbosify(ifile_progress, "Opening %s for reading...\n", data_file);
  ifilter_read_idata(DATA);

  fclose(DATA);
  
  best_target = ifile_rate_categories();
  
  if (best_target < 0)
    {
      ifile_verbosify (ifile_quiet, "Bad folder number (%d)\n", best_target);
      ifile_rcvstore (ea_lines, "inbox");
    }
  else
    ifile_rcvstore(ea_lines, EXT_ARRAY_GET(ea_category, char *, best_target));
  
  ifile_close_log();
  
  return 0;
}


/* Function expects a file pointer which points to the beginning of the idata
   file.  Function will store information from data file into the global
   variables ea_category, as_n, num_categories, as_docs, tot_docs, ea_word,
   as_index, as_age and as_nk. 
   
   Note that the subfunction ifilter_process_frequency_line() skips over lines
   that involve words that do not occur in the message (it checks
   as_msg_freq).  This cuts down considerably on RAM/time consumption.
   */
/* written by Jason Rennie <jr6b+@andrew.cmu.edu> and others for ifile */
void
ifilter_read_idata (FILE * DATA)
{
  int words_read, categories_read;
  DMZ_start=clock();
  
  categories_read = ifile_read_category_names (DATA);
  ifile_read_category_word_counts (DATA);
  ifile_read_category_doc_counts (DATA);
  words_read = ifilter_read_word_frequencies (DATA);
  
  DMZ_end = clock();
  ifile_verbosify(ifile_progress,
		  "Read %d categories, %d words.  Elapsed time: %.3f sec\n",
		  categories_read, words_read,
	          ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
  
}


/* Expects a valid file pointer which points to the beginning of the word
   frequencies section of an idata file.  Function will read lines and store
   the gleamed information into global variables.  In the search of
   efficiency the function will skip lines which do not contain information
   about words in the message at hand.
   Returns number of words information was obtained for. */
/* written by Jason Rennie <jr6b+@andrew.cmu.edu> and others for ifile */
int
ifilter_read_word_frequencies (FILE * DATA)
{
  char line[MAX_STR_LEN*4];
  int words_read = 0;

  num_words = 0;

  readline(line, sizeof(line), DATA);
  while (!feof(DATA))
    {
      num_words++;
      words_read += ifilter_process_frequency_line(line);
      readline(line, sizeof(line), DATA);
    }
  return words_read;
}


/* Given a line from an idata file, function determines whether information
   is needed and if so, parses line and stores information in global array.
   Returns number of words information was obtained for. */
/* written by Jason Rennie <jr6b+@andrew.cmu.edu> and others for ifile */
int
ifilter_process_frequency_line (char *line)
{
  int cat_num, freq;
  char cur_word[MAX_STR_LEN];
  char *token = NULL;
  char freq_entry[MAX_STR_LEN];

  token = strtok(line, " \t");
  
  if (token == NULL)
    return 0;
  
  if (assoc_array_int_get(as_msg_freq, token) == 0)
    return 0;
  
  EXT_ARRAY_SET(ea_word, char *, num_words, strdup(token));
  assoc_array_int_set(as_index, token, num_words); 
  strcpy(cur_word, token);
  
  ifile_verbosify(ifile_debug, "%s ", token);
  
  /* Read age of word */
  token = strtok(NULL, " \t");
  assoc_array_int_set(as_age, cur_word, atoi(token)); 

  /* Read frequency per category  cat_num:freq */
  token = strtok(NULL, ":");
  while (token != NULL)
    {	
      cat_num = atoi(token);
      token = strtok(NULL, " \t");	   
      freq = atoi(token);
      sprintf(freq_entry, "%s,%s", 
	      EXT_ARRAY_GET(ea_word, char*, num_words),
	      EXT_ARRAY_GET(ea_category, char*, cat_num));
      assoc_array_int_set(as_nk, freq_entry, freq);
      ifile_verbosify(ifile_debug, "%d:%d ", cat_num, freq); 
      /* sums total frequency per word */
      assoc_array_int_set(as_sum_nk, EXT_ARRAY_GET(ea_word, char *, num_words), assoc_array_int_get(as_sum_nk, EXT_ARRAY_GET(ea_word, char *, num_words)) + freq);
      token = strtok(NULL, ":");
    }
  ifile_verbosify(ifile_debug, "\n");
  return 1;
}


/* Prints information about program usage and aborts execution */
/* Written by Jason Rennie <jr6b+@andrew.cmu.edu> for ifile */
void 
ifilter_print_usage (void)
{
  ifile_close_log();

  fprintf(stderr, "Usage: ifilter < [input_file]\n");
  fprintf(stderr, "  ifilter will read a single message from standard input and will create\n");
  fprintf(stderr, "  a new copy of the message in the mail folder which it sees fit.\n\n");
  fprintf(stderr, "ifile v%d.%d.%d\n", IFILE_MAJOR_VERSION_NUMBER,
          IFILE_MINOR_VERSION_NUMBER, IFILE_TRIFLING_VERSION_NUMBER);
  exit(0);
}




